#!/opt/alt/python27/bin/python
#coding:utf-8



import getopt
import glob
import os
import re
import shutil
import subprocess
import sys
import syslog
from tempfile import mkstemp

INSTALL_AS_NAME = 'httpd24-alt-mod-passenger'

PASSENGER_TEMP_DIR = '/var/passenger'
PASSENGER_DIR = '/opt/passenger-httpd24-5.3.7-9.el7.cloudlinux'
PASSENGER_ARCH = "/usr/share/alt-mod-passenger-httpd24-5.3.7-9.el7.cloudlinux/archives/passenger.tar.gz"
PASSENGER_LIBS_ARCH = "/usr/share/alt-mod-passenger-httpd24-5.3.7-9.el7.cloudlinux/archives/libspassenger.tar.gz"

INTERPRETER_PATH = '/opt/alt/ruby21/bin/ruby'
PASSENGER_MOD_PATH = 'buildout/apache2/mod_passenger.so'
PASSENGER_EXEC_PATH = 'bin/passenger-install-apache2-module'

APACHE_CONFIG = "/opt/rh/httpd24/root/etc/httpd/conf/conf.d/mod_passenger-5.3.7-9.el7.cloudlinux.conf"
CPANEL_APACHE_CONF = "/opt/rh/httpd24/root/etc/httpd/conf/includes/pre_main_global.conf"
CPANEL_EXTRA_APACHE_CONF = APACHE_CONFIG
DA_APACHE_CONF = "/opt/rh/httpd24/root/etc/httpd/conf/extra/httpd-includes.conf"
OTHER_CONF = "/opt/rh/httpd24/root/etc/httpd/conf.d/mod_passenger-5.3.7-9.el7.cloudlinux.conf"
PASSENGER_INCLUDE_CONST = "\n#For passenger\nInclude "
PASSENGER_INCLUDE_PATH = "/opt/alt/alt-mod-passenger/etc/mod_passenger-5.3.7-9.el7.cloudlinux.conf"
PASSENGER_INCLUDE_DIRECTIVE = PASSENGER_INCLUDE_CONST + PASSENGER_INCLUDE_PATH
INSTALL_LOG_PATH = "/var/log/httpd24-alt-mod-passenger_install.log"
VER_ID = "5.3.7-9.el7.cloudlinux"

NEED_BUILD = False

CPNAME = subprocess.check_output(['cldetect', '--detect-cp-name']).strip()
CAGEFSCTL = '/usr/sbin/cagefsctl'
_syslog = lambda x,y: None

class OopsError (RuntimeError): pass


def open_syslog():

    syslog.openlog(INSTALL_AS_NAME, syslog.LOG_PID, syslog.LOG_LOCAL7)
    syslog.syslog(syslog.LOG_INFO, str(sys.argv))
    syslog.syslog(syslog.LOG_INFO, "Control panel detected as '%s'." % CPNAME)

    return syslog.syslog

def get_pattern():
    return re.compile(
        r"""LoadModule\s+passenger_module.*?<IfModule mod_passenger\.c>.*?</IfModule>""",
        re.M|re.S)


def add_passenger_directive(path):
    """
    Add passenger "Include" directive to httpd config file
    """
    if os.path.exists(path):
        with open(path, "r+") as f:
            content = f.read()
            if PASSENGER_INCLUDE_PATH not in content:
                # add include directive if not exists
                info("Append Include directive into '%s'" % (path))
                f.write(PASSENGER_INCLUDE_DIRECTIVE)
        check_obsolete_config_options(os.path.dirname(path))
    else:
        with open(path, "w+") as f:
            info("Set Include directive into '%s'" % (path))
            f.write(PASSENGER_INCLUDE_DIRECTIVE)

    return True


def remove_passenger_directive(path):
    """
    Remove "Include" passenger directive from httpd config file
    """
    if os.path.isfile(path):
        with open(path, "r+") as f:
            fd, tmp_path = mkstemp()
            with open(tmp_path, "w+") as tmp_file:
                for line in f:
                    if PASSENGER_INCLUDE_PATH in line:
                        info("Remove Include directive from '%s'" % (path))
                    elif "For passenger" in line:
                        pass
                    else:
                        tmp_file.write(line)

            # For compatibility with tests
            with open(tmp_path, "r+") as tmp_file:
                data = tmp_file.read()
                data = data.strip()
                tmp_file.seek(0)
                tmp_file.truncate()
                tmp_file.write(data)

        os.remove(path)
        shutil.move(tmp_path, path)


def clear_folder(path):
    """
    analog of `rm -rf <path>`
    """

    def err_report(f,p,e): sys.stderr.write('%s\n' % e[1])

    if os.path.exists(path):
        for item in glob.glob('%s/*' % path):
            if os.path.isdir(item):
                shutil.rmtree(item, ignore_errors=False, onerror=err_report)
            else:
                try: os.unlink(item)
                except EnvironmentError as e:
                    sys.stderr.write('%s\n' % e)
##
## Commented cz tests are failed on (os.mkdir(path))
##   OSError: [Errno 17] File exists: '/tmp/tmpqt0hq4'
##
#    if os.path.exists(path):
#        shutil.rmtree(path, ignore_errors=False, onerror=lambda f,p,e: sys.stderr.write('%s\n' % e[1]))
#        os.mkdir(path)


def inflate(path, arch_path):
    """
    Extract passenger archive (sources or pre-builded)
    """
    clear_folder(path)
    cmd = ['/bin/tar', '-zxf', arch_path, '-C', path, '--strip-components', '1']

    returncode = subprocess.call(cmd)
    if returncode:
        raise OopsError("Cannot inflate [%d] : %s" % (returncode, ' '.join(cmd)))

def err(msg):
    sys.stderr.write("\033[41mERROR: {}\033[00m\n" .format(msg))
    syslog.syslog(syslog.LOG_ERR, "ERROR: %s" % msg)
def warn(msg):
    sys.stderr.write("\033[91mWARNING: {}\033[00m\n" .format(msg))
    syslog.syslog(syslog.LOG_WARNING, "WARNING: %s" % msg)
def info(msg):
    sys.stderr.write("\033[94mINFO: {}\033[00m\n" .format(msg))
    syslog.syslog(syslog.LOG_INFO, "INFO: %s" % msg)

def check_obsolete_config_options(path):
    """
    Scans config files to check for obsolete options
    The passenger+apache will complain for presence of such options
    """
    if os.path.isfile(path):
        with open(path, "r") as f:
            content = f.read()
            for opt in ['PassengerTempDir']:
                if opt in content:
                    warn('`%s`: found obsolete option `%s`\n' % (path, opt))
    if os.path.isdir(path):
        for item in glob.glob('%s/*.conf' % path):
            check_obsolete_config_options(item)

def add_to_apache():
    """
    Add directives to apache config for each panel
    """
    if CPNAME == "cPanel":
        patt = get_pattern()
        if os.path.exists(CPANEL_APACHE_CONF):
            with open(CPANEL_APACHE_CONF, 'r+') as f:
                data = f.read()
                blocks = patt.findall(data)
                if len(blocks):
                    # remove old blocks from config
                    data = patt.sub("", data).strip()
                    f.seek(0)
                    f.truncate()
                    f.write(data)
                    info("Remove old Load module directive from '%s'" % (CPANEL_APACHE_CONF))

            add_passenger_directive(CPANEL_APACHE_CONF)
            check_obsolete_config_options(os.path.dirname(CPANEL_APACHE_CONF))

        if os.path.exists(os.path.dirname(CPANEL_EXTRA_APACHE_CONF)):
            add_passenger_directive(CPANEL_EXTRA_APACHE_CONF)

    elif CPNAME == "DirectAdmin":
        add_passenger_directive(DA_APACHE_CONF)

    else:
        # add for non-panel server default httpd config
        conf_dir = os.path.dirname(OTHER_CONF)
        if not os.path.exists(conf_dir):
            os.makedirs(conf_dir, 0755)
            os.chmod(conf_dir, 0755)

        add_passenger_directive(OTHER_CONF)


def delete_from_apache():
    """
    Remove passenger directives from apache configs for each panel
    """
    if CPNAME == "cPanel":
        patt = get_pattern()
        if os.path.exists(CPANEL_APACHE_CONF):
            with open(CPANEL_APACHE_CONF, "r+") as f:
                data = f.read()
                # remove old cpanel blocks
                new_data = re.sub(patt, '', data).strip()

                if data != new_data:
                    # rewrite config only if changed
                    f.seek(0)
                    f.truncate()
                    f.write(new_data)
                    info("Remove cPanel Include & Load directive from '%s'" % (CPANEL_APACHE_CONF))

            remove_passenger_directive(CPANEL_APACHE_CONF)

        remove_passenger_directive(CPANEL_EXTRA_APACHE_CONF)

    elif CPNAME == "DirectAdmin":
        remove_passenger_directive(DA_APACHE_CONF)

    else:
        remove_passenger_directive(OTHER_CONF)


def restart_apache():
    cmd = ['/sbin/service', 'httpd24-httpd', 'restart']
    returncode = subprocess.call(cmd, stdout=open('/dev/null','a'),
                                 stderr=open('/dev/null','a'))
    if returncode:
        raise OopsError("Cannot restart apache [%d] : %s" % (returncode, ' '.join(cmd)))


def install():
    """
    Build passenger from sources and install apache directives for each panel
    """

    try:
        os.mkdir(PASSENGER_DIR)
        os.mkdir(os.path.dirname(APACHE_CONFIG))
    except OSError:
        pass

    if NEED_BUILD and CPNAME in ["cPanel", "DirectAdmin"]:
        # for this panels build module from sources
        inflate(PASSENGER_DIR, PASSENGER_ARCH)
        os.chdir(PASSENGER_DIR)
        passenger_script_path = '/usr/share/alt-mod-passenger-httpd24-5.3.7-9.el7.cloudlinux/utils/passenger'
        if os.path.exists(passenger_script_path) and os.path.isfile(passenger_script_path):
            install_log = open(INSTALL_LOG_PATH, 'a')
            cmd = [passenger_script_path]
            try:
                info("Build passenger...")
                subprocess.check_call(
                    cmd,
                    stdout=install_log,
                    stderr=install_log)
                info("Build passenger... OK")
            except subprocess.CalledProcessError:
                raise SystemExit("Could not compile module with [%s]. For details see [%s]" % (passenger_script_path, install_log,))

    else:
        # extract pre-builded files
        inflate(PASSENGER_DIR, PASSENGER_LIBS_ARCH)

    passenger_mod_exe = "%s/%s" % (PASSENGER_DIR, PASSENGER_MOD_PATH, )
#    if CPNAME == 'cPanel':
#        passenger_mod_exe_cpanel = "/usr/lib64/apache2/modules/mod_passenger.so"
#        shutil.copy2(passenger_mod_exe, passenger_mod_exe_cpanel)
#        passenger_mod_exe = passenger_mod_exe_cpanel
    if not os.path.exists(passenger_mod_exe):
        raise OopsError("Could not find module %s" % (passenger_mod_exe, ))

    add_to_apache()
    #setup_passenger_cagefs() <- done via rpm %triggerin scriptlet
    #restart_apache()


def uninstall():
    """
    Uninstall passenger configurations from system files and remove
    builded/extracted targets
    """
    delete_from_apache()
    clear_folder(PASSENGER_DIR)


def setup_passenger_cagefs():
    """
    Setup passenger module into cagefs
    """
    if not os.path.exists(CAGEFSCTL):
        return
    try:
        cagefs_setup_mount_dir_passenger(PASSENGER_TEMP_DIR)
    except EnvironmentError as e:
        raise OopsError("Could not setup Passenger [%s]" % str(e))


def cagefs_setup_mount_dir_passenger(dir_path, mode=0o755):
    """
    Add CageFS mount point for passenger directories
    :param dir_path: directory path
    :param mode: permissions
    """
    # ensure permissions
    os.chmod(dir_path, mode)
    # add new line if needed
    subprocess.call([CAGEFSCTL, "--check-mp"])
    cagefs_mp = open('/etc/cagefs/cagefs.mp', "r+")
    try:
        # check if dir is already present
        if not filter(lambda x: x.startswith(dir_path), cagefs_mp):
            # add dir
            cagefs_mp.seek(0, 2)  # 2: seek to the end of file
            print >> cagefs_mp, "# next line is added by alt-mod-passenger"
            print >> cagefs_mp, dir_path
    finally:
        cagefs_mp.close()
    # remount
    subprocess.call([CAGEFSCTL, "--remount-all"])


def check_file(file_s, file, CreateLinks = True):
    """
    File 'file_s' is the name which system is operated with and expect to be present.
    File 'file' is unique versioned name of 'file_s'.
    THis function checks and links them togeter.

    Checks file 'file_s' for presence and creates (erases) it as symbolic link to 'file':
      file_s -> file

    """

#    if not os.path.exists(file_s) and CreateLinks:
#        warn("The link file '%s' is absent!\n" % (file_s, ))

    if os.path.exists(file_s):
        os.unlink(file_s)

#    dest_file = "%s/%s" % (os.path.dirname(file), file_s)
    dest_file = os.path.basename(file)

    if CreateLinks:
        if not os.path.exists(file):
            err("The destination file '%s' is absent!\n" % (file, ))
        else:
            cmd = ['/bin/ln', '-s', '-f', dest_file, file_s]
            returncode = subprocess.call(cmd, stdout=open('/dev/null','a'),
                                              stderr=open('/dev/null','a'))
            if returncode:
                err("Cannot make link [%d]: '%s'" % (returncode, ' '.join(cmd)))
            else:
                info("Made link file '%s' -> '%s'!\n" % (file_s, dest_file))
#    else:
#        if os.path.exists(dest_file):
#            os.unlink(dest_file)


def fix_links(CreateLinks = True):
    """
    Checks and makes/creates sym-links to versioned resources
    """

#    check_file("/opt/alt/alt-mod-passenger/etc/mod_passenger-5.3.7-9.el7.cloudlinux.conf", "mod_passenger.conf", CreateLinks)
    check_file("/etc/profile.d/alt_mod_passenger.sh", "/etc/profile.d/alt_mod_passenger.sh-5.3.7-9.el7.cloudlinux", CreateLinks)

    if CPNAME == "cPanel":
        pass
#        check_file(CPANEL_APACHE_CONF, CreateLinks)
#        check_file(CPANEL_EXTRA_APACHE_CONF, CreateLinks)

    elif CPNAME == "DirectAdmin":
        pass
#        check_file(DA_APACHE_CONF, CreateLinks)

    else:
        # add for non-panel server the default httpd config
#        check_file(OTHER_CONF, "mod_passenger.conf", CreateLinks)
        pass

def usage():
    print 'Usage: %s OPTIONS' % (sys.argv[0],)
    print ''
    print '-i| --install : installs the module'
    print '-d| --delete  : removes the module'
    print '-h| --help    : shows this help'
    print '-p| --path    : path to look for sources. Default path is %s' % (PASSENGER_DIR,)
    print '-c| --config  : path to apache conf file. Default is %s' % (CPANEL_EXTRA_APACHE_CONF,)
    print '--fix-links   : fix symbolic links to resources. Create if absent.'
    print '--del-links   : delete symbolic links to resources. Does nothing if absent.'
    print '--setup-passenger-cagefs : configure passenger with CageFS'


def main(argv):
    """
    Main run function
    """
    opts, args = getopt.getopt(argv, 
        'idhp:c:',
        ['install', 'delete', 'path=', 'config=', 'help', "setup-passenger-cagefs", 'fix-links', 'del-links'])

    global _syslog
    _syslog = open_syslog()

    for o, a in opts:
        if o in ['-h', '--help']:
            usage()
            sys.exit(0)
        elif o in ['-i', '--install']:
            install()
        elif o in ['-d', '--delete']:
            uninstall()
        elif o in ['-p', '--path']:
            PASSENGER_DIR = a
        elif o in ['-c', '--config']:
            APACHE_CONFIG = a
        elif o == '--setup-passenger-cagefs':
            setup_passenger_cagefs()
        elif o == '--fix-links':
            fix_links()
        elif o == '--del-links':
            fix_links(CreateLinks=False)


if __name__ == '__main__':

    try:
        main(sys.argv[1:])
    except OopsError as e:
        _syslog(syslog.LOG_ERR, str(e))
        sys.exit(1)
    except Exception as e:
        _syslog(syslog.LOG_ERR, str(e))
        # dump backtrace to stderr for unexpected exception type
        raise
    else:
        _syslog(syslog.LOG_INFO, "No errors.")
        sys.exit(0)
