#!/opt/alt/python27/bin/python
#coding:utf-8



from glob import glob
from collections import Counter
import os
from subprocess import Popen, PIPE
import sys
import syslog


# save base path for execute scripts
BASE_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
# wrapper for get full path
REL = lambda x: os.path.join(BASE_PATH, x)

_syslog = lambda x,y: None

PASSENGER_CONFIG_BIN = "/opt/passenger-httpd24-5.3.7-9.el7.cloudlinux/bin/passenger-config"
PASSENGER_BUILD_NATIVE_SUPPORT_PARAM = "build-native-support"
PASSENGER_ABOUT_NATIVE_SUPPORT_CMD = "about"
PASSENGER_ABOUT_NATIVE_SUPPORT_PARAM = "--rubyext-compat-id"
PASSENGER_NATIVE_SUPPORT_LIB_PATTERN = \
    '/opt/passenger-httpd24-5.3.7-9.el7.cloudlinux/buildout/ruby/ruby-[123]*/passenger_native_support.so'

DEFAULT_RUBY = '/usr/bin/ruby'


class ExecError (RuntimeError): pass


def execute(*cmd, **kwargs):
    """
    Execute cmd wrapper
    """
    p = Popen(cmd, stdout=PIPE, stderr=PIPE)
    stdout, stderr = p.communicate()
    if p.returncode:
        raise ExecError(stdout +  stderr)
    return stdout


def install_cpanel():
    """
    Install scripts for cpanel
    """
    # to locate /usr/local/apache/bin/apachectl from easyapache
    os.environ["PATH"] = "/usr/local/apache/bin:%s" % os.environ["PATH"]

    execute(REL("utils/cpanel-make-lve"))
    execute(REL("utils/cpanel-hooks-lve"))
    execute(REL("hooks/cpanelpostupcp.hook"))


def install_directadmin():
    """
    Install configs to directadmin panel
    """
    other_install()
    # restart apache


def other_install():
    """
    Install module for default system
    """
    # better than use exec call
    from handle_module import install
    install()


def delete_cpanel():
    """
    Delete scripts for cpanel
    """
    execute(REL("utils/cpanel-delete-lve"))


def delete_directadmin():
    """
    Delete actions for directadmin
    """
    other_delete()
    # restart apache


def other_delete():
    """
    Remove passenger for default system
    """
    # better than use exec call
    from handle_module import uninstall
    uninstall()

class MakeRubyNativeLib:

    RC_PRE_BUILT = 'PRE-BUILT'
    RC_BUILD_OK  = 'BUILD OK'
    RC_BUILD_ERR = 'BUILD ERROR'
    RC_NO_RUBY   = 'NO RUBY'

    @staticmethod
    def check_passenger_config_bin():

        # alternative construction to
        # if not os.path.exists:
        #   raise "Oops! Error"
        try:
            open(PASSENGER_CONFIG_BIN, 'rb')
        except EnvironmentError as e:
            raise RuntimeError("Malconfigured alt-mod-passenger setup : %s" % e)

    @staticmethod
    def check_all_ruby_packages_installed(ruby_ver):
        for pack in ['devel', 'rubygems']:
            try:
                pack_name = "alt-%s-%s" % (ruby_ver, pack)
                cmd = ['rpm', '-q', pack_name]
                execute(*cmd)

            except ExecError as e:
                mess = 'Absent package %s, required to build native libs!\nIn order to use %s with Passenger, please install %s!' % (pack_name, ruby_ver, pack_name)
                _syslog(syslog.LOG_ERR, '%s\n' % mess)
                sys.stderr.write('%s\n' % mess)
                return False

        return True

    @classmethod
    def make_ruby_native_lib(cls, ruby_cmd):
        """
        Builds native support library for particular ruby version (default if None)
        """

        cls.check_passenger_config_bin()

        if os.path.exists(ruby_cmd):

            pre_built_list = glob(PASSENGER_NATIVE_SUPPORT_LIB_PATTERN)
            try:
                cmd = [ruby_cmd, PASSENGER_CONFIG_BIN, PASSENGER_BUILD_NATIVE_SUPPORT_PARAM]
                execute(*cmd)

                cmd = [ruby_cmd,
                       PASSENGER_CONFIG_BIN, PASSENGER_ABOUT_NATIVE_SUPPORT_CMD,
                       PASSENGER_ABOUT_NATIVE_SUPPORT_PARAM]
                atom = execute(*cmd).strip()
            except ExecError as e:
                _syslog(syslog.LOG_ERR, '%s : %s' % (cmd, str(e)))
                return cls.RC_BUILD_ERR

            lib = "/opt/passenger-httpd24-5.3.7-9.el7.cloudlinux/buildout/ruby/%s/passenger_native_support.so" % (atom, )
            if os.path.exists(lib):
                if lib in pre_built_list:
                    return cls.RC_PRE_BUILT
                else:
                    return cls.RC_BUILD_OK
            else:
                return cls.RC_BUILD_ERR
        else:
            # It is ok if there is no ruby: no ruby - no native lib for this ruby version
            return cls.RC_NO_RUBY

def build_ruby_native_support():
    """
    Build ruby native support library
    """

    impl = MakeRubyNativeLib # mk alias
    res = Counter()

    # tune default ruby version
    if os.path.exists(DEFAULT_RUBY):
        res.update([ impl.make_ruby_native_lib(DEFAULT_RUBY) ])

    # find ruby alt versions
    for f in os.listdir('/opt/alt'):
        ruby_cmd ="/opt/alt/%s/bin/ruby" % f
        if f.startswith('ruby') and os.path.exists(ruby_cmd) and impl.check_all_ruby_packages_installed(f):
            # invoke tune command
            res.update([ impl.make_ruby_native_lib(ruby_cmd) ])

    def _count(seq): return sum(1 for _ in seq)
    expect_ok = _count(rc for rc in res.elements() if rc != impl.RC_NO_RUBY)
    rc_ok = _count(rc for rc in res.elements() if rc in [impl.RC_BUILD_OK,
                                                         impl.RC_PRE_BUILT])

    if res[impl.RC_BUILD_ERR]:
        oops_msg = "Cannot build complete set of ruby native support libs: %d/%d" \
                 % (rc_ok, expect_ok)
        _syslog(syslog.LOG_ERR, oops_msg)
        sys.stderr.write('%s\n' % oops_msg)

    elif res[impl.RC_BUILD_OK]:
        msg = "Have just built complete set of ruby native support libs: %d/%d" \
                % (rc_ok, expect_ok)
        _syslog(syslog.LOG_INFO, msg)
        sys.stderr.write('%s\n' % msg)

def main(args):
    """
    Module instal scripts
    """
    from handle_module import open_syslog
    global _syslog

    CPNAME = execute("cldetect", '--detect-cp-name').strip()
    if "-i" in args or "--install" in args:
        _syslog = open_syslog()

        if CPNAME == "cPanel":
            install_cpanel()
        elif CPNAME == "DirectAdmin":
            install_directadmin()
        else:
            other_install()

    elif "-d" in args or "--delete" in args:
        _syslog = open_syslog()

        if CPNAME == "cPanel":
            delete_cpanel()
        elif CPNAME == "DirectAdmin":
            delete_directadmin()
        else:
            other_delete()

    elif "--build-ruby-native-support" in args:
        _syslog = open_syslog()

        build_ruby_native_support()

if "__main__" == __name__:

    try:
        exit_code = main(sys.argv[1:])
    except Exception as e:
        _syslog(syslog.LOG_ERR, str(e))
        # dump backtrace to stderr
        raise
    else:
        _syslog(syslog.LOG_INFO, "No errors.")
